import os
from pkg_resources import parse_version

_ROOT = os.path.abspath(os.path.dirname(__file__))
__version__ = "4.0.0a10"
VERSION = parse_version(__version__)


def get_data_dir():
    return os.path.join(_ROOT, 'data')



##
##

import logging
import os
from enum import Enum
from typing import Union, List
from Crypto.PublicKey import RSA
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
import hashlib
from cryptography.exceptions import UnsupportedAlgorithm
from couchformation.exception import FatalError, NonFatalError

logger = logging.getLogger('couchformation.ssh')
logger.addHandler(logging.NullHandler())
HOME_DIRECTORY = os.path.expanduser('~')

SSH_PATHS = [
    HOME_DIRECTORY + '/.ssh',
    HOME_DIRECTORY,
    HOME_DIRECTORY + '/Documents',
    HOME_DIRECTORY + '/Downloads'
]


class SSHExtensions(Enum):
    PEM = ".pem"
    DER = ".der"
    KEY = ".key"


class SSHError(FatalError):
    pass


class EmptyResultSet(NonFatalError):
    pass


class SSHUtil(object):

    def __init__(self):
        pass

    @staticmethod
    def ssh_key_absolute_path(name: str) -> Union[str, None]:
        for location in SSH_PATHS:
            for file_found in os.listdir(location):
                if file_found == name or next((f"{name}{e.value}" for e in SSHExtensions if f"{name}{e.value}" == file_found), None):
                    return location + '/' + file_found
        return None

    @staticmethod
    def list_private_key_files() -> Union[List[dict], None]:
        dir_list = []
        key_file_list = []

        for location in SSH_PATHS:
            for file_name in os.listdir(location):
                full_path = location + '/' + file_name
                dir_list.append(full_path)

        for i in range(len(dir_list)):
            if not os.path.isfile(dir_list[i]) or not os.access(dir_list[i], os.R_OK):
                continue

            file_ext = os.path.splitext(dir_list[i])
            if file_ext[0] != "id_rsa":
                if not next((e for e in SSHExtensions if e.value == file_ext[1]), None):
                    continue

            file_handle = open(dir_list[i], 'r')
            blob = file_handle.read()
            pem_key_bytes = str.encode(blob)

            try:
                key = serialization.load_pem_private_key(
                    pem_key_bytes, password=None, backend=default_backend()
                )
            except (ValueError, TypeError, UnsupportedAlgorithm):
                continue

            key_entry = {"file": dir_list[i]}
            pri_der = key.private_bytes(
                encoding=serialization.Encoding.DER,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )
            der_digest = hashlib.sha1(pri_der)
            hex_digest = der_digest.hexdigest()
            key_fingerprint = ':'.join(hex_digest[i:i + 2] for i in range(0, len(hex_digest), 2))
            key_entry.update({"fingerprint": key_fingerprint})
            pub_der = key.public_key().public_bytes(
                serialization.Encoding.DER,
                serialization.PublicFormat.SubjectPublicKeyInfo
            )
            pub_der_digest = hashlib.md5(pub_der)
            pub_hex_digest = pub_der_digest.hexdigest()
            pub_key_fingerprint = ':'.join(pub_hex_digest[i:i + 2] for i in range(0, len(pub_hex_digest), 2))
            key_entry.update({"pub_fingerprint": pub_key_fingerprint})

            key_file_list.append(key_entry)

        if len(key_file_list) == 0:
            raise EmptyResultSet("No SSH keys found. Please make sure you have at least one SSH key configured.")

        return key_file_list

    def get_key_by_fingerprint(self, fingerprint: str):
        key_list = self.list_private_key_files()
        for key in key_list:
            if key['fingerprint'] == fingerprint or key['pub_fingerprint'] == fingerprint:
                return key['file']
        return None

    @staticmethod
    def get_ssh_public_key(key_file: str) -> str:
        if not os.path.isabs(key_file):
            key_file = SSHUtil.ssh_key_absolute_path(key_file)
        fh = open(key_file, 'r')
        key_pem = fh.read()
        fh.close()
        rsa_key = RSA.importKey(key_pem)
        modulus = rsa_key.n
        pub_exp_e = rsa_key.e
        pri_exp_d = rsa_key.d
        prime_p = rsa_key.p
        prime_q = rsa_key.q
        private_key = RSA.construct((modulus, pub_exp_e, pri_exp_d, prime_p, prime_q))
        public_key = private_key.public_key().exportKey('OpenSSH')
        ssh_public_key = public_key.decode('utf-8')
        return ssh_public_key

    @staticmethod
    def write_file(file_name: str, data: str) -> bool:
        try:
            file_handle = open(file_name, 'w')
            file_handle.write(data)
            file_handle.write("\n")
            file_handle.close()
            return True
        except OSError as err:
            raise SSHError(f"can not write to file {file_name}: {err}.")

    @staticmethod
    def get_ssh_public_key_file(key_file: str) -> str:
        dir_list = []

        if not os.path.isabs(key_file):
            key_file = SSHUtil.ssh_key_absolute_path(key_file)

        gen_public_key = SSHUtil.get_ssh_public_key(key_file)

        for location in SSH_PATHS:
            for file_name in os.listdir(location):
                full_path = location + '/' + file_name
                dir_list.append(full_path)

        for i in range(len(dir_list)):
            if not os.path.isfile(dir_list[i]) or not os.access(dir_list[i], os.R_OK):
                continue
            file_handle = open(dir_list[i], 'r')
            try:
                public_key = file_handle.readline()
            except (UnicodeDecodeError, IOError):
                continue
            file_size = os.fstat(file_handle.fileno()).st_size
            read_size = len(public_key)
            if file_size != read_size:
                continue
            public_key = public_key.rstrip()
            key_parts = public_key.split(' ')
            pub_key_part = ' '.join(key_parts[0:2])
            pub_key_bytes = str.encode(pub_key_part)
            try:
                serialization.load_ssh_public_key(pub_key_bytes)
            except (ValueError, UnsupportedAlgorithm):
                continue
            if gen_public_key == public_key:
                return dir_list[i]

        private_key_dir = os.path.dirname(key_file)
        private_key_file = os.path.basename(key_file)
        private_key_name = os.path.splitext(private_key_file)[0]
        pub_file_name = private_key_dir + '/' + private_key_name + '.pub'

        SSHUtil.write_file(pub_file_name, gen_public_key)

        return pub_file_name



##
##

from __future__ import annotations
import os
import attr
import json
from pwd import getpwnam
from grp import getgrnam
from typing import Optional, List, Union

INFRASTRUCTURE = 0x01
INSTANCES = 0x02


class StateError(Exception):
    pass


@attr.s
class AWSInstance:
    name: Optional[str] = attr.ib(default=None)
    instance_id: Optional[str] = attr.ib(default=None)
    machine_type: Optional[str] = attr.ib(default=None)
    volume_iops: Optional[str] = attr.ib(default="3000")
    volume_size: Optional[str] = attr.ib(default="256")
    volume_type: Optional[str] = attr.ib(default="gp3")
    root_iops: Optional[str] = attr.ib(default="3000")
    root_size: Optional[str] = attr.ib(default="256")
    root_type: Optional[str] = attr.ib(default="gp3")
    public_ip: Optional[str] = attr.ib(default=None)
    private_ip: Optional[str] = attr.ib(default=None)
    services: Optional[str] = attr.ib(default=None)
    zone: Optional[str] = attr.ib(default=None)
    subnet_id: Optional[str] = attr.ib(default=None)


@attr.s
class AWSInstanceSet:
    cloud: Optional[str] = attr.ib(default="aws")
    name: Optional[str] = attr.ib(default=None)
    username: Optional[str] = attr.ib(default=None)
    instance_list: Optional[List[dict]] = attr.ib(default=[])


@attr.s
class AWSZone:
    zone: Optional[str] = attr.ib(default=None)
    subnet_id: Optional[str] = attr.ib(default=None)
    cidr: Optional[str] = attr.ib(default=None)


@attr.s
class AWSState:
    cloud: Optional[str] = attr.ib(default="aws")
    region: Optional[str] = attr.ib(default=None)
    vpc_id: Optional[str] = attr.ib(default=None)
    vpc_cidr: Optional[str] = attr.ib(default=None)
    security_group_id: Optional[str] = attr.ib(default=None)
    ssh_key: Optional[str] = attr.ib(default=None)
    internet_gateway_id: Optional[str] = attr.ib(default=None)
    route_table_id: Optional[str] = attr.ib(default=None)
    zone_list: Optional[List[dict]] = attr.ib(default=[])


@attr.s
class GCPDisk:
    name: Optional[str] = attr.ib(default=None)
    zone: Optional[str] = attr.ib(default=None)


@attr.s
class GCPInstance:
    name: Optional[str] = attr.ib(default=None)
    machine_type: Optional[str] = attr.ib(default=None)
    volume_size: Optional[str] = attr.ib(default="256")
    volume_type: Optional[str] = attr.ib(default="pd-ssd")
    root_size: Optional[str] = attr.ib(default="256")
    root_type: Optional[str] = attr.ib(default="pd-ssd")
    public_ip: Optional[str] = attr.ib(default=None)
    private_ip: Optional[str] = attr.ib(default=None)
    services: Optional[str] = attr.ib(default=None)
    zone: Optional[str] = attr.ib(default=None)
    disk_list: Optional[List[dict]] = attr.ib(default=[])


@attr.s
class GCPInstanceSet:
    cloud: Optional[str] = attr.ib(default="gcp")
    name: Optional[str] = attr.ib(default=None)
    username: Optional[str] = attr.ib(default=None)
    instance_list: Optional[List[dict]] = attr.ib(default=[])


@attr.s
class GCPZone:
    zone: Optional[str] = attr.ib(default=None)
    subnet: Optional[str] = attr.ib(default=None)


@attr.s
class GCPState:
    cloud: Optional[str] = attr.ib(default="gcp")
    region: Optional[str] = attr.ib(default=None)
    network: Optional[str] = attr.ib(default=None)
    network_cidr: Optional[str] = attr.ib(default=None)
    subnet: Optional[str] = attr.ib(default=None)
    subnet_cidr: Optional[str] = attr.ib(default=None)
    ssh_key: Optional[str] = attr.ib(default=None)
    firewall_default: Optional[str] = attr.ib(default=None)
    firewall_cbs: Optional[str] = attr.ib(default=None)
    firewall_ssh: Optional[str] = attr.ib(default=None)
    gcp_project: Optional[str] = attr.ib(default=None)
    credentials: Optional[str] = attr.ib(default=None)
    zone_list: Optional[List[dict]] = attr.ib(default=[])


@attr.s
class AzureDisk:
    name: Optional[str] = attr.ib(default=None)
    zone: Optional[str] = attr.ib(default=None)
    disk_attachment:  Optional[str] = attr.ib(default=None)


@attr.s
class AzureInstance:
    name: Optional[str] = attr.ib(default=None)
    node_nic: Optional[str] = attr.ib(default=None)
    node_pub_ip: Optional[str] = attr.ib(default=None)
    resource_group: Optional[str] = attr.ib(default=None)
    machine_type: Optional[str] = attr.ib(default=None)
    volume_tier: Optional[str] = attr.ib(default="P20")
    volume_size: Optional[str] = attr.ib(default="256")
    volume_type: Optional[str] = attr.ib(default="Premium_LRS")
    root_tier: Optional[str] = attr.ib(default="P20")
    root_size: Optional[str] = attr.ib(default="256")
    root_type: Optional[str] = attr.ib(default="Premium_LRS")
    public_ip: Optional[str] = attr.ib(default=None)
    private_ip: Optional[str] = attr.ib(default=None)
    services: Optional[str] = attr.ib(default=None)
    zone: Optional[str] = attr.ib(default=None)
    vm_public_ip: Optional[str] = attr.ib(default=None)
    vm_nic: Optional[str] = attr.ib(default=None)
    vm_nsg_association: Optional[str] = attr.ib(default=None)
    disk_list: Optional[List[dict]] = attr.ib(default=[])


@attr.s
class AzureInstanceSet:
    cloud: Optional[str] = attr.ib(default="azure")
    name: Optional[str] = attr.ib(default=None)
    username: Optional[str] = attr.ib(default=None)
    instance_list: Optional[List[dict]] = attr.ib(default=[])


@attr.s
class AzureZone:
    zone: Optional[str] = attr.ib(default=None)
    subnet: Optional[str] = attr.ib(default=None)
    subnet_id: Optional[str] = attr.ib(default=None)


@attr.s
class AzureState:
    cloud: Optional[str] = attr.ib(default="azure")
    location: Optional[str] = attr.ib(default=None)
    resource_group: Optional[str] = attr.ib(default=None)
    network: Optional[str] = attr.ib(default=None)
    network_id: Optional[str] = attr.ib(default=None)
    network_cidr: Optional[str] = attr.ib(default=None)
    subnet: Optional[str] = attr.ib(default=None)
    subnet_id: Optional[str] = attr.ib(default=None)
    subnet_cidr: Optional[str] = attr.ib(default=None)
    ssh_key: Optional[str] = attr.ib(default=None)
    network_security_group: Optional[str] = attr.ib(default=None)
    network_security_group_id: Optional[str] = attr.ib(default=None)
    zone_list: Optional[List[dict]] = attr.ib(default=[])


@attr.s
class ServiceSet:
    services: Optional[dict] = attr.ib(default={})

    def import_list(self, services: dict):
        for name, node_list in services.items():
            private_ip_list = [ip for ip in node_list.list_private_ip()]
            public_ip_list = [ip for ip in node_list.list_public_ip()]
            self.services.update({
                name: {
                    'private': private_ip_list,
                    'public': public_ip_list
                }
            })


@attr.s
class StateConfig:
    name: Optional[str] = attr.ib(default=None)
    cloud: Optional[str] = attr.ib(default=None)
    project_dir: Optional[str] = attr.ib(default=None)

    def set(self, name, cloud, project_dir):
        self.name = name
        self.cloud = cloud
        self.project_dir = project_dir


config = StateConfig()
services = ServiceSet()

infrastructure: Union[AWSState, GCPState, AzureState] = AWSState()
_infrastructure_update = False
instance_set: Union[AWSInstanceSet, GCPInstanceSet, AzureInstanceSet] = AWSInstanceSet()
_instance_update = False


def read_file(name: str):
    try:
        with open(name, 'r') as cfg_file_h:
            data = json.load(cfg_file_h)
            return data
    except FileNotFoundError:
        return None
    except Exception as err:
        raise StateError(f"can not read from config file {name}: {err}")


def write_file(data: dict, name: str):
    try:
        with open(name, 'w') as cfg_file_h:
            json.dump(data, cfg_file_h, indent=2)
            cfg_file_h.write('\n')
    except Exception as err:
        raise StateError(f"can not write to config file {name}: {err}")


def make_dir(name: str, owner: str = None, group: str = None, mode: int = 0o775):
    owner_id = getpwnam(owner).pw_uid if owner else None
    group_id = getgrnam(group).gr_gid if group else None
    if not os.path.exists(name):
        path_dir = os.path.dirname(name)
        if not os.path.exists(path_dir):
            make_dir(path_dir)
        try:
            uid = os.stat(path_dir).st_uid if not owner_id else owner_id
            gid = os.stat(path_dir).st_gid if not group_id else group_id
            os.mkdir(name)
            os.chown(name, uid, gid)
            os.chmod(name, mode)
        except OSError:
            raise


def service_dir():
    return os.path.join(config.project_dir, config.name)


def switch_cloud() -> None:
    global infrastructure, instance_set
    common_state_data = {}
    resource_state_data = {}
    common_state_file = os.path.join(config.project_dir, 'state.json')
    resource_state_file = os.path.join(service_dir(), 'state.json')

    if os.path.exists(common_state_file):
        common_state_data = read_file(common_state_file)
        if common_state_data['cloud'] != config.cloud:
            raise StateError(f"Cloud mismatch: state: {common_state_data['cloud']} requested: {config.cloud}")
    if os.path.exists(resource_state_file):
        resource_state_data = read_file(resource_state_file)
        if resource_state_data['cloud'] != config.cloud:
            raise StateError(f"Cloud mismatch: state: {resource_state_data['cloud']} requested: {config.cloud}")

    if config.cloud == 'aws':
        infrastructure = AWSState(**common_state_data)
        instance_set = AWSInstanceSet(**resource_state_data)
    elif config.cloud == 'gcp':
        infrastructure = GCPState(**common_state_data)
        instance_set = GCPInstanceSet(**resource_state_data)
    elif config.cloud == 'azure':
        infrastructure = AzureState(**common_state_data)
        instance_set = AzureInstanceSet(**resource_state_data)


def update(mask):
    global _infrastructure_update, _instance_update
    if 0x01 & mask > 0:
        _infrastructure_update = True
    if 0x02 & mask > 0:
        _instance_update = True


def save():
    if _infrastructure_update:
        make_dir(config.project_dir)
        # noinspection PyTypeChecker
        write_file(attr.asdict(infrastructure), os.path.join(config.project_dir, 'state.json'))
    if _instance_update:
        make_dir(service_dir())
        # noinspection PyTypeChecker
        write_file(attr.asdict(instance_set), os.path.join(service_dir(), 'state.json'))


def instance_count():
    return len(instance_set.instance_list)


def services_active():
    deployment_file = os.path.join(config.project_dir, "deployment.cfg")
    if os.path.exists(deployment_file):
        with open(deployment_file, 'r') as cfg_file:
            deployment_data = json.load(cfg_file)
            for name, service in deployment_data.items():
                if name == 'core':
                    continue
                state_path = os.path.join(config.project_dir, name, 'state.json')
                if os.path.exists(state_path):
                    with open(state_path, 'r') as state_file:
                        state_data = json.load(state_file)
                        if state_data.get('instance_list') and len(state_data.get('instance_list')) > 0:
                            return True
    return False


def infrastructure_display():
    json.dumps(infrastructure, indent=2)


def instances_display():
    json.dumps(instance_set, indent=2)



##
##

import attr
import os
import json
import argparse
import couchformation.constants as C
from typing import Optional, List, Tuple
from enum import Enum
from couchformation.exception import FatalError
from couchformation.config import BaseConfig, NodeConfig, Parameters, AuthMode, get_project_dir
from couchformation.util import FileManager, dict_merge
from couchformation.kvdb import KeyValueStore

DEPLOYMENT = "deployment.db"


class DeploymentError(FatalError):
    pass


class ServiceModel(str, Enum):
    cbs = 'cbs'
    sgw = 'sgw'
    sdk = 'sdk'
    generic = 'generic'


@attr.s
class Service:
    cloud: Optional[str] = attr.ib(default="aws")
    name: Optional[str] = attr.ib(default="nodes")
    model: Optional[str] = attr.ib(default=ServiceModel.generic)
    region: Optional[str] = attr.ib(default=None)
    os_id: Optional[str] = attr.ib(default=None)
    os_version: Optional[str] = attr.ib(default=None)
    auth_mode: Optional[str] = attr.ib(default="default")
    profile: Optional[str] = attr.ib(default='default')
    connect_svc: Optional[str] = attr.ib(default=None)
    connect_ip: Optional[str] = attr.ib(default=None)
    config: Optional[List[NodeConfig]] = attr.ib(default=[])

    def from_dict(self, options: dict):
        for attribute in self.__annotations__:
            if attribute == 'config':
                continue
            if options.get(attribute):
                setattr(self, attribute, options.get(attribute))

    @property
    def auth(self) -> AuthMode:
        return AuthMode[self.auth_mode]


class NodeGroup(object):

    def __init__(self, options: argparse.Namespace):
        self.options = options
        self.project = self.options.project
        self.name = self.options.name
        self.cloud = self.options.cloud

        filename = os.path.join(get_project_dir(self.project), f"{self.name}.db")
        network = os.path.join(get_project_dir(self.project), C.NETWORK)
        metadata = os.path.join(get_project_dir(self.project), C.METADATA)
        self.db = KeyValueStore(filename)
        self.net = KeyValueStore(network)
        self.meta = KeyValueStore(metadata)

    def create_network(self, parameters: argparse.Namespace, group=1):
        document = f"network:{self.cloud}"

        opt_dict = vars(self.options)
        parm_dict = vars(parameters)
        combined = dict_merge(opt_dict, parm_dict)

        if group == 1:
            self.net.remove(document)
            self.net.document(document)
            self.meta.document('network')
            self.net.update(combined)
            self.meta[self.cloud] = True

    def create_node_group(self, parameters: argparse.Namespace, group=1):
        document = f"{self.name}:{group:04d}"

        opt_dict = vars(self.options)
        parm_dict = vars(parameters)
        combined = dict_merge(opt_dict, parm_dict)

        if group == 1:
            self.db.clean()
        self.db.document(document)
        self.meta.document('resources')
        self.db.update(combined)
        self.meta[self.name] = self.cloud

        self.create_network(parameters, group)

    def add_to_node_group(self, parameters: argparse.Namespace):
        count = len(self.db.doc_id_startswith(self.name))
        if count == 0:
            raise ValueError(f"attempting to add to empty node group")
        self.create_node_group(parameters, count + 1)

    def get_node_groups(self) -> List[KeyValueStore]:
        self.meta.document('resources')
        for resource in self.meta.keys():
            filename = os.path.join(get_project_dir(self.project), f"{resource}.db")
            db = KeyValueStore(filename)
            doc_list = db.doc_id_startswith(resource)
            yield [KeyValueStore(filename, doc) for doc in doc_list]

    def get_networks(self) -> List[KeyValueStore]:
        doc_list = self.net.doc_id_startswith('network')
        return [KeyValueStore(self.net.file_name, doc) for doc in doc_list]


class Deployment(object):

    def __init__(self, parameters: Parameters):
        self.parameters = parameters
        self.core, self.deploy_set = self.from_file()

        try:
            FileManager().make_dir(self.parameters.project_dir)
        except Exception as err:
            raise DeploymentError(f"can not create working dir: {err}")

    @property
    def services(self) -> Tuple[str, BaseConfig, Service]:
        for name, service in self.deploy_set.items():
            yield name, self.core, service

    def store_config(self, overwrite: bool = False):
        service = self.parameters.name
        if not self.deploy_set.get(service) or overwrite:
            new_svc = Service(config=[])
            new_svc.from_dict(self.parameters.as_dict)
            self.deploy_set.update({service: new_svc})
        config = NodeConfig().create(self.parameters.as_dict)
        self.deploy_set.get(service).config.append(config)

    def from_file(self):
        deploy_set = {}
        core = BaseConfig()
        config_data = self.read_file("deployment.cfg")

        if config_data.get('core'):
            core.from_dict(config_data.get('core'))

        core.from_dict(self.parameters.as_dict)

        for name, service in config_data.items():
            if name == 'core':
                continue
            new_svc = Service(config=[])
            new_svc.from_dict(service)
            for saved_config in service.get('config'):
                config = NodeConfig().create(saved_config)
                new_svc.config.append(config)
            deploy_set.update({name: new_svc})

        return core, deploy_set

    def to_file(self):
        config_data = {
            'core': self.core.as_dict
        }
        for name, service in self.deploy_set.items():
            # noinspection PyTypeChecker
            config_data.update({name: attr.asdict(service)})
        self.write_file(config_data, "deployment.cfg")

    def read_file(self, name: str):
        cfg_file = os.path.join(self.parameters.project_dir, name)
        try:
            with open(cfg_file, 'r') as cfg_file_h:
                data = json.load(cfg_file_h)
                return data
        except FileNotFoundError:
            return {}
        except Exception as err:
            raise DeploymentError(f"can not read from config file {cfg_file}: {err}")

    def write_file(self, data: dict, name: str):
        cfg_file = os.path.join(self.parameters.project_dir, name)
        try:
            with open(cfg_file, 'w') as cfg_file_h:
                json.dump(data, cfg_file_h, indent=2)
                cfg_file_h.write('\n')
        except Exception as err:
            raise DeploymentError(f"can not write to config file {cfg_file}: {err}")



##
##

import os
from couchformation import get_data_dir

ROOT_DIRECTORY = os.path.join(os.environ.get('HOME'), '.config', 'couch-formation')
STATE_DIRECTORY = os.path.join(ROOT_DIRECTORY, 'state')
LOG_DIRECTORY = os.path.join(ROOT_DIRECTORY, 'log')
DATA_DIRECTORY = get_data_dir()
NODE_PROFILES = os.path.join(DATA_DIRECTORY, "node_profiles.yaml")
TARGET_PROFILES = os.path.join(DATA_DIRECTORY, "target_profiles.yaml")
PROVISIONER_PROFILES = os.path.join(DATA_DIRECTORY, "provisioner_profiles.yaml")
PLAYBOOK_DIR = os.path.join(DATA_DIRECTORY, "playbooks")

METADATA = "metadata.db"
NETWORK = "network.db"
STATE = "state.db"

GREY_COLOR = "\x1b[38;20m"
YELLOW_COLOR = "\x1b[33;20m"
RED_COLOR = "\x1b[31;20m"
BOLD_RED_COLOR = "\x1b[31;1m"
GREEN_COLOR = "\x1b[32;20m"
SCREEN_RESET = "\x1b[0m"
FORMAT_LEVEL = "%(levelname)s"
FORMAT_NAME = "%(name)s"
FORMAT_MESSAGE = "%(message)s"
FORMAT_LINE = "(%(filename)s:%(lineno)d)"
FORMAT_EXTRA = " [%(name)s](%(filename)s:%(lineno)d)"
FORMAT_TIMESTAMP = "%(asctime)s"

OS_VERSION_LIST = {
    'amzn': ['2', '2023'],
    'rhel': ['8', '9'],
    'centos': ['8'],
    'ol': ['8', '9'],
    'rocky': ['8', '9'],
    'fedora': ['34'],
    'sles': ['12', '15'],
    'opensuse-leap': ['15'],
    'ubuntu': ['20.04', '22.04'],
    'debian': ['10', '11'],
    'arch': []
}

MACHINE_TYPES = [
        {
            "name": "2x4",
            "cpu": 2,
            "memory": 4096
        },
        {
            "name": "2x8",
            "cpu": 2,
            "memory": 8192
        },
        {
            "name": "4x8",
            "cpu": 4,
            "memory": 8192
        },
        {
            "name": "4x16",
            "cpu": 4,
            "memory": 16384
        },
        {
            "name": "8x16",
            "cpu": 8,
            "memory": 16384
        },
        {
            "name": "8x32",
            "cpu": 8,
            "memory": 32768
        },
        {
            "name": "16x32",
            "cpu": 16,
            "memory": 32768
        },
        {
            "name": "16x64",
            "cpu": 16,
            "memory": 65536
        },
        {
            "name": "32x64",
            "cpu": 32,
            "memory": 65536
        },
        {
            "name": "32x128",
            "cpu": 32,
            "memory": 131072
        },
        {
            "name": "48x192",
            "cpu": 48,
            "memory": 196608
        },
        {
            "name": "64x128",
            "cpu": 64,
            "memory": 131072
        },
        {
            "name": "64x256",
            "cpu": 64,
            "memory": 262144
        },
        {
            "name": "80x320",
            "cpu": 80,
            "memory": 327680
        },
        {
            "name": "2x16",
            "cpu": 2,
            "memory": 16384
        },
        {
            "name": "4x32",
            "cpu": 4,
            "memory": 32768
        },
        {
            "name": "8x64",
            "cpu": 8,
            "memory": 65536
        },
        {
            "name": "16x128",
            "cpu": 16,
            "memory": 131072
        },
        {
            "name": "32x256",
            "cpu": 32,
            "memory": 262144
        },
        {
            "name": "48x384",
            "cpu": 48,
            "memory": 393216
        },
        {
            "name": "64x512",
            "cpu": 64,
            "memory": 524288
        },
        {
            "name": "80x640",
            "cpu": 80,
            "memory": 655360
        },
        {
            "name": "2x2",
            "cpu": 2,
            "memory": 2048
        },
        {
            "name": "4x4",
            "cpu": 4,
            "memory": 4096
        },
        {
            "name": "8x8",
            "cpu": 8,
            "memory": 8192
        },
        {
            "name": "16x16",
            "cpu": 16,
            "memory": 16384
        },
        {
            "name": "32x32",
            "cpu": 32,
            "memory": 32768
        },
        {
            "name": "48x48",
            "cpu": 48,
            "memory": 49152
        },
        {
            "name": "64x64",
            "cpu": 64,
            "memory": 65536
        },
        {
            "name": "80x80",
            "cpu": 80,
            "memory": 81920
        }
    ]

provisioners = {
    "cbs": {
        "pre_provision": [
            'curl -sfL https://raw.githubusercontent.com/mminichino/host-prep-lib/main/bin/setup.sh | sudo -E bash -s - -s -g https://github.com/mminichino/host-prep-lib',
        ],
        "provision": [
            'sudo bundlemgr -b CBS',
            'sudo swmgr cluster create -n testdb -s {{ SERVICES }} -g {{ NODE_ZONE }} -D /cbdata -l {{ PRIVATE_IP_LIST }}',
        ],
        "post_provision": [
            'sudo swmgr cluster rebalance -l {{ PRIVATE_IP_LIST }}',
        ]
    },
    "sgw": {
        "pre_provision": [
            'curl -sfL https://raw.githubusercontent.com/mminichino/host-prep-lib/main/bin/setup.sh | sudo -E bash -s - -s -g https://github.com/mminichino/host-prep-lib',
        ],
        "provision": [
            'sudo bundlemgr -b InstallSGW',
            'sudo swmgr gateway configure -l {{ CONNECT_LIST }}',
            'sudo bundlemgr -b EnableSGW'
        ],
        "post_provision": []
    }
}



##
##

import logging
import warnings
import argparse
from overrides import override
from couchformation.cli.cli import CLI
import couchformation.kvdb as kvdb

warnings.filterwarnings("ignore")
logger = logging.getLogger()


class DBDumpCLI(CLI):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @override()
    def local_args(self):
        opt_parser = argparse.ArgumentParser(parents=[self.parser], add_help=False)
        opt_parser.add_argument('-t', '--table', action='store', help="Table Name")

    def run(self):
        for file in self.remainder:
            for doc in kvdb.documents(file):
                print(f"Document: {doc.document_id}")
                for key, value in doc.items():
                    print(f"{key:<12} = {value}")


def main(args=None):
    cli = DBDumpCLI(args)
    cli.run()



##
##

import os
import logging
import uuid
import collections
from uuid import UUID
from shutil import copyfile
from pwd import getpwnam
from grp import getgrnam
from couchformation.exception import FatalError

logger = logging.getLogger('couchformation.util')
logger.addHandler(logging.NullHandler())


class FileManagerError(FatalError):
    pass


def dict_merge(dict1, dict2):
    new_dict = dict(dict1)
    new_dict.update(dict2)
    return new_dict


def dict_merge_list(*dicts):
    res = collections.defaultdict(list)
    for d in dicts:
        for k, v in d.items():
            res[f"{k}_list"].append(v)
    return {k: set(v) for k, v in res.items()}


class FileManager(object):

    def __init__(self):
        pass

    def make_dir(self, name: str, owner: str = None, group: str = None, mode: int = 0o775):
        owner_id = getpwnam(owner).pw_uid if owner else None
        group_id = getgrnam(group).gr_gid if group else None
        if not os.path.exists(name):
            path_dir = os.path.dirname(name)
            if not os.path.exists(path_dir):
                self.make_dir(path_dir)
            try:
                uid = os.stat(path_dir).st_uid if not owner_id else owner_id
                gid = os.stat(path_dir).st_gid if not group_id else group_id
                os.mkdir(name)
                os.chown(name, uid, gid)
                os.chmod(name, mode)
            except OSError:
                raise

    def set_perms(self, name: str, owner: str, group: str, mode: int = 0o775):
        if os.path.exists(name):
            uid = getpwnam(owner).pw_uid
            gid = getgrnam(group).gr_gid
            os.chown(name, uid, gid)
            os.chmod(name, mode)
        else:
            self.make_dir(name, owner, group, mode)

    @staticmethod
    def copy_file(source: str, destination: str) -> None:
        try:
            logger.debug(f"Copying {source} to {destination}")
            copyfile(source, destination)
        except Exception as err:
            raise FileManagerError(f"can not copy {source} to {destination}: {err}")


class UUIDGen(object):

    def __init__(self):
        self._last_uuid = UUID('00000000000000000000000000000000')
        self._uuid = self._last_uuid

    def recompute(self, text):
        self._uuid = uuid.uuid5(self._last_uuid, text)
        self._last_uuid = self._uuid

    @property
    def uuid(self):
        return self._uuid



##
##

import os
import json
import logging
import sqlite3
import tempfile
from collections import UserDict

logger = logging.getLogger('couchformation.kvdb')
logger.addHandler(logging.NullHandler())


def connect(*args, **kwargs):
    return KeyValueStore(*args, **kwargs)


def documents(*args, **kwargs):
    return KeyValueStore(*args, **kwargs).documents()


class KeyValueStore(UserDict):

    def __init__(self, filename=None, tablename="kv"):
        super().__init__()
        self.temp_file = filename is None
        if filename:
            self.filename = filename
        else:
            f, self.filename = tempfile.mkstemp(prefix='kv_dict')
            os.close(f)
        self.tablename = tablename
        self.conn, self.cursor = self._connect()
        self._table(self.tablename)

    def _table(self, tablename):
        self.conn.execute(f"""CREATE TABLE IF NOT EXISTS \"{tablename}\" (key TEXT PRIMARY KEY, value BLOB)""")

    def _connect(self):
        try:
            conn = sqlite3.connect(self.filename)
            cursor = conn.cursor()
            return conn, cursor
        except Exception as err:
            message = f"Can not initialize connection for file {self.filename}: {err}"
            logger.debug(message)
            raise RuntimeError(message)

    def _select(self, query, arg=None):
        if not arg:
            arg = ()
        self.cursor.execute(query, arg)
        return self.cursor.fetchall()

    def __enter__(self):
        if not hasattr(self, 'conn') or self.conn is None:
            self.conn, self.cursor = self._connect()
        return self

    def __exit__(self, *exc_info):
        self.close()

    def __str__(self):
        return f"KeyValueStore({self.filename}, {self.tablename})"

    def __repr__(self):
        return str(self)

    def __len__(self):
        rows = self._select(f"""SELECT COUNT(*) FROM \"{self.tablename}\"""")[0]
        return rows[0] if rows is not None else 0

    def __bool__(self):
        rows = self._select(f"""SELECT MAX(ROWID) FROM \"{self.tablename}\"""")[0]
        return True if rows[0] is not None else False

    def iterkeys(self):
        for row in self._select(f"""SELECT key FROM \"{self.tablename}\" ORDER BY rowid"""):
            yield row[0]

    def itervalues(self):
        for row in self._select(f"""SELECT value FROM \"{self.tablename}\" ORDER BY rowid"""):
            yield row[0]

    def iteritems(self):
        for row in self._select(f"""SELECT key, value FROM \"{self.tablename}\" ORDER BY rowid"""):
            yield row[0], row[1]

    def keys(self):
        return self.iterkeys()

    def values(self):
        return self.itervalues()

    def items(self):
        return self.iteritems()

    def __contains__(self, key):
        return self._select(f"""SELECT 1 FROM \"{self.tablename}\" WHERE key = ?""", (key,)) is not None

    def __getitem__(self, key):
        item = self._select(f"""SELECT value FROM \"{self.tablename}\" WHERE key = ?""", (key,))
        if len(item) == 0:
            return None
        return item[0][0]

    def __setitem__(self, key, value):
        self.conn.execute(f"""REPLACE INTO \"{self.tablename}\" (key, value) VALUES (?,?)""", (key, value))
        self.commit()

    def __delitem__(self, key):
        if key not in self:
            return
        self.conn.execute(f"""DELETE FROM \"{self.tablename}\" WHERE key = ?""", (key,))
        self.commit()

    @property
    def as_dict(self):
        d = {}
        for k, v in self.iteritems():
            d.update({k: v})
        return d

    def list_add(self, name, *args):
        data = self.list_get(name)
        data.append(args)
        new_list = json.dumps(data)
        self.conn.execute(f"""REPLACE INTO \"{self.tablename}\" (key, value) VALUES (?,?)""", (name, new_list))
        self.commit()

    def list_remove(self, name, match):
        data = self.list_get(name)
        new_data = [e for e in data if match not in e]
        new_list = json.dumps(new_data)
        self.conn.execute(f"""REPLACE INTO \"{self.tablename}\" (key, value) VALUES (?,?)""", (name, new_list))
        self.commit()

    def list_get(self, name):
        item = self._select(f"""SELECT value FROM \"{self.tablename}\" WHERE key = ?""", (name,))
        if len(item) == 0:
            current_list = '[]'
        else:
            current_list = item[0][0]
        data = json.loads(current_list)
        return data

    def list_element(self, name, match):
        data = self.list_get(name)
        for element in data:
            if match in element:
                return element
        return None

    def list_exists(self, name, match):
        data = self.list_get(name)
        for element in data:
            if match in element:
                return True
        return False

    def list_len(self, name):
        data = self.list_get(name)
        return len(data)

    def update(self, *args, **kwargs):
        if not args and not kwargs:
            return
        items = [(k, v) for k, v in kwargs.items()]
        for d in args:
            items.extend([(k, v) for k, v in d.items()])

        self.conn.executemany(f"""REPLACE INTO \"{self.tablename}\" (key, value) VALUES (?, ?)""", items)
        self.commit()

    def __iter__(self):
        return self.iterkeys()

    def clear(self):
        self.conn.commit()
        self.conn.execute(f"""DELETE FROM \"{self.tablename}\";""")
        self.conn.commit()

    def remove(self, name):
        self.conn.commit()
        try:
            self.conn.execute(f"""DROP TABLE \"{name}\";""")
        except sqlite3.OperationalError:
            pass
        self.conn.commit()

    def clean(self):
        self.cursor.execute("""SELECT name FROM sqlite_master WHERE type=\"table\"""")
        res = self.cursor.fetchall()
        for name in res:
            self.conn.commit()
            self.conn.execute(f"""DROP TABLE \"{name[0]}\";""")
            self.conn.commit()

    @staticmethod
    def get_document_names(filename):
        if not os.path.isfile(filename):
            raise IOError(f"file {filename} does not exist")
        with sqlite3.connect(filename) as conn:
            cursor = conn.execute("""SELECT name FROM sqlite_master WHERE type=\"table\"""")
            res = cursor.fetchall()
        return [name[0] for name in res]

    def document_len(self, name):
        rows = self._select(f"""SELECT COUNT(*) FROM \"{name}\"""")[0]
        return rows[0]

    def doc_id_startswith(self, text):
        self.cursor.execute(f"""SELECT name FROM sqlite_master WHERE type=\"table\" AND name LIKE '{text}%'""")
        res = self.cursor.fetchall()
        return [name[0] for name in res]

    def document_exists(self, name):
        try:
            self._select(f"""SELECT COUNT(*) FROM \"{name}\"""")
            return True
        except sqlite3.OperationalError:
            return False

    @property
    def document_id(self):
        return self.tablename

    @property
    def file_name(self):
        return self.filename

    def document(self, name):
        self.tablename = name
        self._table(self.tablename)

    def documents(self):
        self.cursor.execute("""SELECT name FROM sqlite_master WHERE type=\"table\"""")
        res = self.cursor.fetchall()
        return [KeyValueStore(self.filename, name[0]) for name in res if self.document_len(name[0]) > 0]

    def commit(self):
        if self.conn is not None:
            self.conn.commit()

    def close(self):
        logger.debug(f"closing {self}")
        if hasattr(self, 'conn') and self.conn is not None:
            self.conn.commit()
            self.conn.close()
            self.conn = None
        if self.temp_file:
            try:
                os.remove(self.filename)
            except Exception as err:
                logger.debug(f"temp file delete: {err}")
                pass

    def terminate(self):
        self.close()
        if self.filename == ':memory:':
            return
        logger.info(f"deleting {self.filename}")
        try:
            if os.path.isfile(self.filename):
                os.remove(self.filename)
        except (OSError, IOError):
            logger.exception(f"failed to delete {self.filename}")

    def __del__(self):
        try:
            self.close()
        except Exception as err:
            logger.debug(f"deleting object: {err}")
            pass



##
##

import logging
from couchformation.exception import FatalError
from couchformation.aws.node import AWSDeployment
from couchformation.gcp.node import GCPDeployment
from couchformation.azure.node import AzureDeployment
from couchformation.config import get_project_dir
from couchformation.deployment import NodeGroup
from couchformation.executor.targets import TargetProfile, ProvisionerProfile, BuildProfile
from couchformation.executor.dispatch import JobDispatch

logger = logging.getLogger('couchformation.exec.process')
logger.addHandler(logging.NullHandler())


class ProjectError(FatalError):
    pass


class Project(object):

    def __init__(self, args, remainder):
        self.options = args
        self.remainder = remainder
        self.cloud = self.options.cloud
        self.provisioner = self.options.provisioner
        self.runner = JobDispatch()

    @staticmethod
    def deployer(cloud: str):
        if cloud == 'aws':
            return AWSDeployment
        elif cloud == 'gcp':
            return GCPDeployment
        elif cloud == 'azure':
            return AzureDeployment
        else:
            raise ValueError(f"cloud {cloud} is not supported")

    def create(self):
        logger.info(f"Creating new service")
        profile = TargetProfile(self.remainder).get(self.cloud)
        NodeGroup(self.options).create_node_group(profile.options)

    def add(self):
        logger.info(f"Adding node group to service")
        profile = TargetProfile(self.remainder).get(self.cloud)
        NodeGroup(self.options).add_to_node_group(profile.options)

    def deploy(self):
        for net in NodeGroup(self.options).get_networks():
            logger.info(f"Deploying network for cloud {net.get('cloud')}")
            cloud = net.get('cloud')
            profile = TargetProfile(self.remainder).get(cloud)
            module = profile.network.driver
            instance = profile.network.module
            method = profile.network.deploy
            self.runner.foreground(profile.base.driver, profile.base.module, profile.base.test, net.as_dict)
            self.runner.dispatch(module, instance, method, net.as_dict)
        list(self.runner.join())
        for groups in NodeGroup(self.options).get_node_groups():
            number = 0
            for db in groups:
                cloud = db.get('cloud')
                profile = TargetProfile(self.remainder).get(cloud)
                module = profile.node.driver
                instance = profile.node.module
                method = profile.node.deploy
                for n in range(int(db['quantity'])):
                    number += 1
                    logger.info(f"Deploying service {db.get('name')} node group {db.get('group')} node {number}")
                    parameters = db.as_dict
                    parameters['number'] = number
                    self.runner.dispatch(module, instance, method, parameters)
            result_list = list(self.runner.join())
            if len(result_list) != number:
                raise ProjectError(f"Partial deployment: deployed {len(result_list)} expected {number}")
            result_list = sorted(result_list, key=lambda d: d['name'])
            private_ip_list = [d['private_ip'] for d in result_list]
            public_ip_list = [d['public_ip'] for d in result_list]
            result_list = [dict(item, private_ip_list=private_ip_list, public_ip_list=public_ip_list) for item in result_list]

            provisioner = ProvisionerProfile().get(self.provisioner)
            p_module = provisioner.driver
            p_instance = provisioner.module
            p_method = provisioner.method

            p_list = [provisioner.parameter_gen(result, groups[0].as_dict) for result in result_list]

            default = BuildProfile().get('default')

            for step, command in enumerate(default.commands):
                for p_set in p_list:
                    logger.info(f"Provisioning node {p_set.get('name')} - default step #{step + 1}")
                    self.runner.dispatch(p_module, p_instance, p_method, p_set, command, default.root)
                exit_codes = list(self.runner.join())
                if any(n != 0 for n in exit_codes):
                    raise ProjectError(f"Provisioning step failed")

            build = BuildProfile().get(self.options.build)

            for step, command in enumerate(build.commands):
                for p_set in p_list:
                    logger.info(f"Provisioning node {p_set.get('name')} - build step #{step + 1}")
                    self.runner.dispatch(p_module, p_instance, p_method, p_set, command, build.root)
                exit_codes = list(self.runner.join())
                if any(n != 0 for n in exit_codes):
                    raise ProjectError(f"Provisioning step failed")

    def destroy(self):
        for groups in NodeGroup(self.options).get_node_groups():
            number = 0
            for db in groups:
                cloud = db.get('cloud')
                profile = TargetProfile(self.remainder).get(cloud)
                module = profile.node.driver
                instance = profile.node.module
                method = profile.node.destroy
                self.runner.foreground(profile.base.driver, profile.base.module, profile.base.test, db.as_dict)
                for n in range(int(db['quantity'])):
                    number += 1
                    logger.info(f"Removing service {db.get('name')} node group {db.get('group')} node {number}")
                    parameters = db.as_dict
                    parameters['number'] = number
                    self.runner.dispatch(module, instance, method, parameters)
            list(self.runner.join())
        for net in NodeGroup(self.options).get_networks():
            logger.info(f"Removing network for cloud {net.get('cloud')}")
            cloud = net.get('cloud')
            profile = TargetProfile(self.remainder).get(cloud)
            module = profile.network.driver
            instance = profile.network.module
            method = profile.network.destroy
            self.runner.dispatch(module, instance, method, net.as_dict)
        list(self.runner.join())

    def list(self, api=False):
        return_list = []
        for groups in NodeGroup(self.options).get_node_groups():
            number = 0
            for db in groups:
                cloud = db.get('cloud')
                profile = TargetProfile(self.remainder).get(cloud)
                module = profile.node.driver
                instance = profile.node.module
                method = profile.node.info
                for n in range(int(db['quantity'])):
                    number += 1
                    parameters = db.as_dict
                    parameters['number'] = number
                    self.runner.dispatch(module, instance, method, parameters)
            result_list = list(self.runner.join())
            result_list = sorted(result_list, key=lambda d: d['name'])
            for result in result_list:
                if not api:
                    logger.info(f"Node: {result.get('name')} "
                                f"Private IP: {result.get('private_ip'):<15} "
                                f"Public IP: {result.get('public_ip'):<15} "
                                f"Services: {result.get('services')}")
                return_list.append(result)
        return return_list

    @property
    def location(self):
        return get_project_dir(self.options.project)



##
##

import sys
import os
import inspect
import logging
import couchformation.state as state

logger = logging.getLogger('couchformation.error')
logger.addHandler(logging.NullHandler())


class FatalError(Exception):

    def __init__(self, message):
        import traceback
        logging.debug(traceback.print_exc())
        frame = inspect.currentframe().f_back
        (filename, line, function, lines, index) = inspect.getframeinfo(frame)
        filename = os.path.basename(filename)
        logging.debug("Error: {} in {} {} at line {}: {}".format(type(self).__name__, filename, function, line, message))
        logging.error(f"{message} [{filename}:{line}]")
        state.save()
        sys.exit(1)


class NonFatalError(Exception):

    def __init__(self, message):
        frame = inspect.currentframe().f_back
        (filename, line, function, lines, index) = inspect.getframeinfo(frame)
        filename = os.path.basename(filename)
        self.message = "Error: {} in {} {} at line {}: {}".format(
            type(self).__name__, filename, function, line, message)
        super().__init__(self.message)



##
##

import logging
import warnings
import argparse
import sys
import os
import signal
import inspect
import traceback
import datetime
from datetime import datetime
from couchformation.config import get_log_dir
import couchformation.constants as C
from couchformation.util import FileManager
from couchformation.exception import FatalError

warnings.filterwarnings("ignore")
logger = logging.getLogger()


def break_signal_handler(signum, frame):
    signal_name = signal.Signals(signum).name
    (filename, line, function, lines, index) = inspect.getframeinfo(frame)
    logger.info(f"received signal {signal_name} in {filename} {function} at line {line}")
    tb = ''.join(traceback.format_stack(frame))
    logger.info(tb)
    print("")
    print("Break received, aborting.")
    sys.exit(1)


class CloudMgrError(FatalError):
    pass


class CustomDisplayFormatter(logging.Formatter):
    FORMATS = {
        logging.DEBUG: f"[{C.GREY_COLOR}{C.FORMAT_LEVEL}{C.SCREEN_RESET}] {C.FORMAT_MESSAGE}",
        logging.INFO: f"[{C.GREEN_COLOR}{C.FORMAT_LEVEL}{C.SCREEN_RESET}] {C.FORMAT_MESSAGE}",
        logging.WARNING: f"[{C.YELLOW_COLOR}{C.FORMAT_LEVEL}{C.SCREEN_RESET}] {C.FORMAT_MESSAGE}",
        logging.ERROR: f"[{C.RED_COLOR}{C.FORMAT_LEVEL}{C.SCREEN_RESET}] {C.FORMAT_MESSAGE}",
        logging.CRITICAL: f"[{C.BOLD_RED_COLOR}{C.FORMAT_LEVEL}{C.SCREEN_RESET}] {C.FORMAT_MESSAGE}"
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        if logging.DEBUG >= logging.root.level:
            log_fmt += C.FORMAT_EXTRA
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class CustomLogFormatter(logging.Formatter):
    FORMATS = {
        logging.DEBUG: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}",
        logging.INFO: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}",
        logging.WARNING: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}",
        logging.ERROR: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}",
        logging.CRITICAL: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}"
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        if logging.DEBUG >= logging.root.level:
            log_fmt += C.FORMAT_EXTRA
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class StreamOutputLogger(object):
    def __init__(self, _logger, _level, _file=None):
        self.logger = _logger
        self.level = _level
        if not _file:
            self.file = sys.stdout
        else:
            self.file = _file
        self.buffer = ''

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self.logger.log(self.level, line.rstrip())

    def __getattr__(self, name):
        return getattr(self.file, name)

    def flush(self):
        pass


class CLI(object):

    def __init__(self, args):
        self.log_dir = get_log_dir()
        signal.signal(signal.SIGINT, break_signal_handler)
        default_debug_file = os.path.join(self.log_dir, f"{os.path.splitext(os.path.basename(sys.argv[0]))[0]}.log")
        debug_file = os.environ.get("COUCH_FORMATION_DEBUG_FILE", default_debug_file)
        self.args = args
        self.parser = None
        self.options = None
        self.remainder = None

        if self.args is None:
            self.args = sys.argv[1:]

        self.init_parser()

        try:
            FileManager().make_dir(self.log_dir)
        except Exception as err:
            raise CloudMgrError(f"can not create working dir: {err}")

        if sys.stdin and sys.stdin.isatty():
            screen_handler = logging.StreamHandler()
            screen_handler.setFormatter(CustomDisplayFormatter())
            logger.addHandler(screen_handler)

        file_handler = logging.FileHandler(debug_file)
        file_handler.setFormatter(CustomLogFormatter())
        logger.addHandler(file_handler)

        logger.setLevel(logging.INFO)

        self.process_args()

    @staticmethod
    def get_timestamp():
        return datetime.utcnow().strftime("%b %d %H:%M:%S")

    def init_parser(self):
        self.parser = argparse.ArgumentParser(add_help=False)
        self.parser.add_argument('-d', '--debug', action='store_true', help="Debug output")
        self.parser.add_argument('-v', '--verbose', action='store_true', help="Verbose output")

    def local_args(self):
        pass

    def process_args(self):
        self.local_args()
        self.options, self.remainder = self.parser.parse_known_args(self.args)
        if self.options.debug:
            logger.setLevel(logging.DEBUG)



from setuptools import setup
import couchformation

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='couchformation',
    version=couchformation.__version__,
    packages=['couchformation', 'couchformation.cli'],
    url='https://github.com/mminichino/couch-formation-core',
    license='Apache License 2.0',
    author='Michael Minichino',
    python_requires='>=3.8',
    entry_points={
        'console_scripts': [
            'cloudmgr = couchformation.cli.cloudmgr:main',
        ]
    },
    install_requires=[
        "attrs==23.1.0",
        "boto3==1.28.11",
        "botocore==1.31.11",
        "cryptography>=36.0.1",
        "dnspython>=2.1.0",
        "google-api-core>=2.4.0",
        "google-api-python-client>=2.34.0",
        "google-auth>=2.3.3",
        "google-auth-httplib2>=0.1.0",
        "googleapis-common-protos>=1.54.0",
        "google-cloud>=0.34.0",
        "google-cloud-compute>=1.6.1",
        "google-cloud-storage>=2.10.0",
        "google-cloud-network-management>=1.5.4",
        "Jinja2>=3.0.3",
        "passlib>=1.7.4",
        "pycryptodome>=3.12.0",
        "pytz>=2021.3",
        "pyvmomi>=8.0.0.1.1",
        "requests>=2.31.0",
        "urllib3>=1.26.7",
        "azure-common>=1.1.28",
        "azure-core>=1.26.1",
        "azure-mgmt-resource>=22.0.0",
        "azure-identity>=1.12.0",
        "azure.mgmt.network>=22.2.0",
        "azure.mgmt.compute>=29.0.0",
        "azure-mgmt-core>=1.3.2",
        "azure.mgmt.storage>=21.0.0",
        "azure.mgmt.subscription>=3.1.1",
        "ply>=3.11",
        "pytest==7.4.0",
        "pytest-asyncio==0.21.1",
        "pytest-rerunfailures==12.0",
        "sqlite-utils~=3.11",
        "paramiko>=3.3.1",
        "overrides>=7.4.0",
        "bumpversion>=0.6.0"
    ],
    author_email='info@unix.us.com',
    description='Couchbase Cloud Automation',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=["couchbase", "devops", "automation"],
    classifiers=[
          "Development Status :: 4 - Beta",
          "License :: OSI Approved :: Apache Software License",
          "Intended Audience :: Developers",
          "Operating System :: OS Independent",
          "Programming Language :: Python",
          "Programming Language :: Python :: 3",
          "Topic :: Database",
          "Topic :: Software Development :: Libraries",
          "Topic :: Software Development :: Libraries :: Python Modules"],
)



##
##

import logging
import warnings
import argparse
from overrides import override
import couchformation
from couchformation.cli.cli import CLI
from couchformation.project import Project

warnings.filterwarnings("ignore")
logger = logging.getLogger()


class CloudMgrCLI(CLI):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @override()
    def local_args(self):
        opt_parser = argparse.ArgumentParser(parents=[self.parser], add_help=False)
        opt_parser.add_argument('-b', '--build', action='store', help="Build Type", default="cbs")
        opt_parser.add_argument('-c', '--cloud', action='store', help="Infrastructure", default="aws")
        opt_parser.add_argument('-p', '--project', action='store', help="Project Name", default="resources")
        opt_parser.add_argument('-n', '--name', action='store', help="Deployment Name", default="nodes")
        opt_parser.add_argument('-x', '--connect', action='store', help="Connection Name", default=None)
        opt_parser.add_argument('-g', '--group', action='store', help="Group Number", default=1, type=int)
        opt_parser.add_argument('-P', '--provisioner', action='store', help="Provisioner Name", default="remote")

        command_subparser = self.parser.add_subparsers(dest='command')
        command_subparser.add_parser('create', help="Create", parents=[opt_parser], add_help=False)
        command_subparser.add_parser('add', help="Add Resource Group", parents=[opt_parser], add_help=False)
        command_subparser.add_parser('deploy', help="Deploy", parents=[opt_parser], add_help=False)
        command_subparser.add_parser('destroy', help="Remove Resources", parents=[opt_parser], add_help=False)
        command_subparser.add_parser('list', help="Display Information", parents=[opt_parser], add_help=False)

    def run(self):
        logger.info(f"Couch Formation v{couchformation.__version__}")

        project = Project(self.options, self.remainder)

        if self.options.command == "create":
            project.create()
        elif self.options.command == "add":
            project.add()
        elif self.options.command == "deploy":
            project.deploy()
        elif self.options.command == "destroy":
            project.destroy()
        elif self.options.command == "list":
            project.list()


def main(args=None):
    cli = CloudMgrCLI(args)
    cli.run()
